$GGAP_VERSION := "1.9.7";
$GGAP_API_VERSION := "1.3";


#############################################################################
##
#V  $GGAP_DATA
##
$GGAP_DATA := rec(
  debug := false,
  original_funcs := rec(),
  next_prompt := [],
  globals := [],
  globals_sent := false,
  api := ["$GGAP_VERSION", "$GGAP_API_VERSION", "$GGAP_RESET", "$GGAP_DATA"],
);


#############################################################################
##
#V  InfoGGAP
##
if not IsBoundGlobal("InfoGGAP") then
  InfoGGAP := NewInfoClass("InfoGGAP");
  SetInfoLevel(InfoGGAP, 10);
fi;


#############################################################################
##
#F  $GGAP_SEND_DATA()
##
$GGAP_SEND_DATA :=
function(arg)
  local s, a;
  s := OutputTextUser();
  WriteAll(s, "@GGAP@v");
  for a in arg do
    if IsString(a) then
      WriteAll(s, a);
    else
      WriteAll(s, String(a));
    fi;
  od;
  WriteAll(s, "@GGAP@e\c");
end;
Add($GGAP_DATA.api, "$GGAP_SEND_DATA");


#############################################################################
##
#F  $GGAP_SEND_RESULT_ERROR(message)
##
$GGAP_SEND_RESULT_ERROR :=
function(message)
  $GGAP_SEND_DATA("result-error:", message);
end;
Add($GGAP_DATA.api, "$GGAP_SEND_RESULT_ERROR");

$GGAP_SEND_ADDED_DELETED :=
function(added, deleted)
  local w;

  if not IsEmpty(added) then
    Print("@GGAP@vglobals-added:");
    for w in added do
      Print(w, "\n");
    od;
    Print("@GGAP@e\c");
  fi;

  if not IsEmpty(deleted) then
    Print("@GGAP@vglobals-deleted:");
    for w in deleted do
      Print(w, "\n");
    od;
    Print("@GGAP@e\c");
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_SEND_ADDED_DELETED");

$GGAP_SEND_GLOBALS_FIRST_TIME :=
function()
  local func, list, i, remove, add;
  $GGAP_DATA.globals_sent := true;

  remove := ["Unbind.", "Unbind[]"];
  add := ["for", "end", "do", "od", "if", "fi"];

  list := List(NamesGVars());
  for i in [1..Length(list)] do
    if not IsBoundGlobal(list[i]) or list[i] in remove then
      Unbind(list[i]);
    fi;
  od;
  Append(list, add);

  $GGAP_SEND_ADDED_DELETED(AsSet(list), []);
  func := ValueGlobal("NamesUserGVars");
  $GGAP_DATA.globals := AsSet(func());
end;
Add($GGAP_DATA.api, "$GGAP_SEND_GLOBALS_FIRST_TIME");

$GGAP_SEND_GLOBALS :=
function()
  local func, current, added, deleted;

  if not $GGAP_DATA.globals_sent then
    $GGAP_SEND_GLOBALS_FIRST_TIME();
    return;
  fi;

  func := ValueGlobal("NamesUserGVars");
  current := AsSet(func());
  added := Difference(current, $GGAP_DATA.globals);
  deleted := Difference($GGAP_DATA.globals, current);
  $GGAP_SEND_ADDED_DELETED(added, deleted);

  $GGAP_DATA.globals := List(current);
end;
Add($GGAP_DATA.api, "$GGAP_SEND_GLOBALS");

$GGAP_RESEND_GLOBALS :=
function()
  $GGAP_DATA.globals_sent := false;
  $GGAP_SEND_GLOBALS();
end;
Add($GGAP_DATA.api, "$GGAP_RESEND_GLOBALS");


#############################################################################
##
#F  $GGAP_EXEC_COMMAND(cmdname, ...)
##
$GGAP_EXEC_COMMAND :=
function(arg)
  local cmd;

  if IsEmpty(arg) then
    $GGAP_SEND_RESULT_ERROR(0, "bad arguments");
    $GGAP_DATA.next_prompt := [""];
    return;
  fi;

  cmd := arg[1];

  if cmd = "run-command" then
    if Length(arg) <> 1 then
      $GGAP_SEND_RESULT_ERROR("bad arguments for run-command");
      $GGAP_DATA.next_prompt := [""];
    else
      $GGAP_DATA.next_prompt := ["@GGAP@voutput:", "@GGAP@e\c"];
    fi;
  elif cmd = "get-globals" then
    $GGAP_SEND_GLOBALS();
    $GGAP_DATA.next_prompt := [""];
  else
    $GGAP_SEND_RESULT_ERROR(Concatenation("bad command '", String(cmd), "'"));
    $GGAP_DATA.next_prompt := [""];
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_EXEC_COMMAND");

#
# gap/screen.g
#

$GGAP_PAGER := function(lines)
  local   formatted,  linepos,  size,  wd,  pl,  count,  i,
          len,  char;

  formatted := false;
  linepos := 1;

  if IsRecord(lines) then
    if IsBound(lines.formatted) then
      formatted := lines.formatted;
    fi;
    if IsBound(lines.start) then
      linepos := lines.start;
    fi;
    lines := lines.lines;
  fi;

  if IsString(lines) then
    lines := SplitString(lines, "\n", "");
  fi;

  size := SizeScreen();
  wd := QuoInt(size[1]+2, 2);
  # really print line without breaking it
  pl := function(l)
    local   r;
    r := 1;
    while r*wd<=Length(l) do
      Print(l{[(r-1)*wd+1..r*wd]}, "\c");
      r := r+1;
    od;
    if (r-1)*wd < Length(l) then
      Print(l{[(r-1)*wd+1..Length(l)]});
    fi;
    Print("\n");
  end;

  if not formatted then
    # cope with overfull lines
    count:=1;
    while count<=Length(lines) do
      if Length(lines[count])>size[1]-2 then
        # find the last blank before this position
        i:=size[1]-2;
        while i>0 and lines[count][i]<>' ' do
          i:=i-1;
        od;
        if i>0 then
          if not IsBound(lines[count+1]) then
            lines[count+1]:="";
          fi;
          lines[count+1]:=Concatenation(
             lines[count]{[i+1..Length(lines[count])]}," ", lines[count+1]);
          lines[count]:=lines[count]{[1..i-1]};
        fi;
      fi;
      count:=count+1;
    od;
  fi;

  for i in [linepos..Length(lines)] do
    pl(lines[i]);
  od;
end;

# $GGAP_SIZE_SCREEN := function(arg)
# end;

Add($GGAP_DATA.api, "$GGAP_PAGER");
# Add($GGAP_DATA.api, "$GGAP_SIZE_SCREEN");


#
# gap/help.g
#

#
# HELP() from lib/helpbase.gi
#

$GGAP_HELP_DATA := rec(
  known_books := [],
  last_topics := [],
);
Add($GGAP_DATA.api, "$GGAP_HELP_DATA");

$GGAP_SEND_BOOK_INFO := function(book)
end;
Add($GGAP_DATA.api, "$GGAP_SEND_BOOK_INFO");

$GGAP_HELP_SIMPLE_STRING := function(str)
  local trans, c, i;

  trans := "";
  i := 1;
  while i <= Length(str) do
    if str[i] in "\003\007" then
      i := i + 1;
      continue;
    fi;
    if str[i] = '\033' then
      i := i + 1;
      if i + 2 <= Length(str) and str[i] = '[' and str[i+2] = 'm' and str[i+1] in "0123456789" then
        i := i + 3;
      elif i + 3 <= Length(str) and str[i] = '[' and str[i+3] = 'm' and str[i+1] in "0123456789" and str[i+2] in "0123456789" then
        i := i + 4;
      fi;
      continue;
    fi;
    Add(trans, str[i]);
    i := i + 1;
  od;

  NormalizeWhitespace(trans);
  return trans;
end;
Add($GGAP_DATA.api, "$GGAP_HELP_SIMPLE_STRING");

$GGAP_ESCAPE_MARKUP := function(str)
  local escaped, c;
  escaped := "";
  for c in str do
    if c = '"' then
      Append(escaped, "&quot;");
    elif c = '<' then
      Append(escaped, "&lt;");
    elif c = '>' then
      Append(escaped, "&gt;");
    elif c = '\'' then
      Append(escaped, "&apos;");
    else
      Add(escaped, c);
    fi;
  od;
  return escaped;
end;
Add($GGAP_DATA.api, "$GGAP_ESCAPE_MARKUP");

$GGAP_HELP_MAKE_URI := function(str)
  if IsString(str) then
    # let ggap parse it
    return Concatenation("gfile:", str);
  else
    return str;
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_HELP_MAKE_URI");

$GGAP_HELP_PRINT_MATCHES := function(entries)
  local e, cnt, len;

  if IsEmpty(entries) then
    return;
  fi;

  if Length(entries) = 1 then
    $GGAP_SEND_DATA("markup:Help: opening <a href=\"", entries[1][2], "\">",
                    $GGAP_ESCAPE_MARKUP(entries[1][1]), "</a>\n");
    $GGAP_SEND_DATA("script:openUrl('", entries[1][2], "')");
    return;
  fi;

  if Length(entries) > 99 then
    Print("Help: too many entries match this topic, displaying first 100 (type ?2 to get match [2])\n");
    entries := entries{[1..99]};
  else
    Print("Help: several entries match this topic (type ?2 to get match [2])\n");
  fi;

  cnt := 1;
  len := LogInt(Length(entries), 10) + 1;
  for e in entries do
    $GGAP_SEND_DATA("markup:<a href=\"", e[2], "\">[", String(cnt, len), "] ",
                    $GGAP_ESCAPE_MARKUP(e[1]), "</a>\n");
    cnt := cnt + 1;
  od;
end;
Add($GGAP_DATA.api, "$GGAP_HELP_PRINT_MATCHES");

$GGAP_HELP_DO_SHOW_MATCHES := function(matches)
  local m, all_books, book, no, e, data, topic, entries, cnt, len;

  matches := List(matches, x -> [HELP_BOOK_INFO(x[1]), x[2]]);
  all_books := AsSet(List(matches, x -> x[1]));

  for book in all_books do
    if not book in $GGAP_HELP_DATA.known_books then
      Add($GGAP_HELP_DATA.known_books, book);
      $GGAP_SEND_BOOK_INFO(book);
    fi;
  od;

  entries := [];
  for m in matches do
    book := m[1];
    no := m[2];
    e := book.entries[no];
    data := HELP_BOOK_HANDLER.(book.handler).HelpData(book, no, "url");
    if data <> fail then
      data := $GGAP_HELP_MAKE_URI(data);
    fi;
    if data <> fail then
      topic := $GGAP_HELP_SIMPLE_STRING(Concatenation(book.bookname, ": ", e[1]));
      Add(entries, [topic, data]);
    fi;
  od;

  $GGAP_HELP_DATA.last_topics := entries;

  if IsEmpty(entries) then
    return false;
  fi;

  $GGAP_HELP_PRINT_MATCHES(entries);
  return true;
end;
Add($GGAP_DATA.api, "$GGAP_HELP_DO_SHOW_MATCHES");


$GGAP_HELP_SHOW_MATCHES :=
function(books, topic, frombegin)
  local exact, match, x, lines;

  # first get lists of exact and other matches
  x := HELP_GET_MATCHES(books, topic, frombegin);
  exact := x[1];
  match := x[2];

  # no topic found
  if 0 = Length(match) and 0 = Length(exact)  then
    Print("Help: no matching entry found\n");
    return false;
  fi;

  # one exact or together one topic found
  if 1 = Length(exact) or (0 = Length(exact) and 1 = Length(match)) then
    if Length(exact) = 0 then exact := match; fi;
    if $GGAP_HELP_DO_SHOW_MATCHES(exact) then
      return true;
    fi;
  fi;

  match := Concatenation(exact, match);
  if not $GGAP_HELP_DO_SHOW_MATCHES(match) then
    Print("Help: no matching entry found\n");
    return false;
  else
    return true;
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_HELP_SHOW_MATCHES");


# Well, it doesn't work. How to get url of the table of contents?
# $GGAP_HELP_SHOW_BOOKS := function()
#   local book, info, url, line;
#   for book in HELP_KNOWN_BOOKS[2] do
#     info := HELP_BOOK_INFO(book[1]);
#     if info = fail then
#       continue;
#     fi;
#
#     url := HELP_BOOK_HANDLER.(info.handler).HelpData(book, no, "url");
#     if data <> fail then
#       data := $GGAP_HELP_MAKE_URI(data);
#     fi;
#
#     url := "http://google.com";
#     line := ReplacedString(book[1], " (not loaded)", "");
#     if not IsEmpty(book[2]) then
#       Append(line, " - ");
#       Append(line, book[2]);
#     fi;
#     $GGAP_SEND_DATA("markup:<a href=\"", url, "\">", $GGAP_ESCAPE_MARKUP(line), "</a>\n");
#   od;
# end;
# Add($GGAP_DATA.api, "$GGAP_HELP_SHOW_BOOKS");


$GGAP_HELP := function(str)
  local origstr, p, book, books, b;

  origstr := ShallowCopy(str);

  # extract the book
  p := Position(str, ':');
  if p <> fail then
    book := str{[1..p-1]};
    str  := str{[p+1..Length(str)]};
  else
    book := "";
  fi;

  # normalizing for search
  book := SIMPLE_STRING(book);
  str := SIMPLE_STRING(str);

  # we check if `book' MATCH_BEGINs some of the available books
  books := Filtered(HELP_KNOWN_BOOKS[1], bn -> MATCH_BEGIN(bn, book));
  if Length(book) > 0 and Length(books) = 0 then
    Print("Help: None of the available books matches (try: '?books').\n");
    return;
  fi;

  # if the topic is empty show the tutorial
  if book = "" and str = "" then
    $GGAP_HELP("Tutorial: Help");
    return;

  elif not IsEmpty(str) and Int(str) <> fail then
    p := Int(str);
    if p < 1 or p > Length($GGAP_HELP_DATA.last_topics) then
      Print("Help: No topic with number ", p, ".\n");
    else
      $GGAP_HELP_PRINT_MATCHES($GGAP_HELP_DATA.last_topics{[p]});
    fi;
    return;

  # XXX
#   # if the subject is 'Welcome to GAP' display a welcome message
#   elif book = "" and str = "welcome to gap"  then
#     add(books, "Welcome to GAP");

  # XXX
#   # if the topic is 'books' display the table of books
#   elif book = "" and str = "books"  then
#     $GGAP_HELP_SHOW_BOOKS();

  # XXX
#   # if the topic is 'chapters' display the table of chapters
#   elif str = "chapters" or book <> "" and str = "" then
#     add(books, "chapters");

  # XXX
#   # if the topic is 'sections' display the table of sections
#   elif str = "sections" then
#     add(books, "sections");

  # if the topic is '?<string>' search the index for any entries for
  # which <string> is a substring (as opposed to an abbreviation)
  elif Length(str) > 0 and str[1] = '?' then
    str := str{[2..Length(str)]};
    NormalizeWhitespace(str);
    $GGAP_HELP_SHOW_MATCHES(books, str, false);

  # search for this topic
  elif $GGAP_HELP_SHOW_MATCHES(books, str, true) then
    ;
  elif origstr in NAMES_SYSTEM_GVARS then
      Print("Help: '", origstr, "' is currently undocumented.\n",
            "      For details, try ?Undocumented Variables\n" );
  elif book = "" and ForAny(HELP_KNOWN_BOOKS[1], bk -> MATCH_BEGIN(bk, str)) then
      Print("Help: Are you looking for a certain book? (Trying '?", origstr,
            ":' ...\n");
      $GGAP_HELP(Concatenation(origstr, ":"));
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_HELP");


#
# gap/last.g
#

$GGAP_BIND_NAMES := function(bind)
  local bind_global, unbind_global, store_global,
        restore_global, store_and_bind;

  Info(InfoGGAP, 3, "$GGAP_BIND_NAMES");

  unbind_global := function(name)
    if IsBoundGlobal(name) then
      if IsReadOnlyGlobal(name) then
        MakeReadWriteGlobal(name);
      fi;
      UnbindGlobal(name);
    fi;
  end;

  store_global := function(name)
    if IsBoundGlobal(name) then
      $GGAP_DATA.original_funcs.(name) := ValueGlobal(name);
    fi;
  end;

  bind_global := function(name, value)
    unbind_global(name);
    BindGlobal(name, value);
  end;

  store_and_bind := function(name, value)
    store_global(name);
    bind_global(name, value);
  end;

  restore_global := function(name)
    unbind_global(name);
    if IsBound($GGAP_DATA.original_funcs.(name)) then
      BindGlobal(name, $GGAP_DATA.original_funcs.(name));
    fi;
  end;

  if bind then
#     store_global("InfoDoPrint");
#     bind_global("InfoDoPrint",
#     function(arglist)
#       Print("@GGAP-INFO@");
#       CallFuncList(Print, arglist);
#       Print("\n@GGAP-INFO-END@");
#     end);

    bind_global("PrintPromptHook",
    function()
      if not IsEmpty($GGAP_DATA.next_prompt) then
        Print($GGAP_DATA.next_prompt[1]);
        Remove($GGAP_DATA.next_prompt, 1);
      else
        $GGAP_SEND_DATA("prompt:", CPROMPT());
      fi;
    end);

    store_and_bind("ColorPrompt", function(setting)
      if setting = true then
        Print("# ColorPrompt() ignored\n");
      fi;
    end);

    store_and_bind("ANSI_COLORS", false);
    store_and_bind("PAGER_BUILTIN", $GGAP_PAGER);
    store_and_bind("PAGER_EXTERNAL", $GGAP_PAGER);
#     store_and_bind("SizeScreen", $GGAP_SIZE_SCREEN);
  else
    unbind_global("PrintPromptHook");
    restore_global("ColorPrompt");
    restore_global("PAGER_BUILTIN");
    restore_global("PAGER_EXTERNAL");
  fi;
end;
Add($GGAP_DATA.api, "$GGAP_BIND_NAMES");


#############################################################################
##
#F  $GGAP_INIT()
##
$GGAP_INIT :=
function()
  Info(InfoGGAP, 3, "$GGAP_INIT");
  $GGAP_SEND_DATA("status:started");
  $GGAP_BIND_NAMES(true);
  $GGAP_SEND_GLOBALS();
end;
Add($GGAP_DATA.api, "$GGAP_INIT");

#############################################################################
##
#F  $GGAP_DEINIT()
##
##  Undo stuff done in $GGAP_INIT
##
$GGAP_DEINIT := function()
  $GGAP_BIND_NAMES(false);
  Info(InfoGGAP, 3, "$GGAP_DEINIT");
end;
Add($GGAP_DATA.api, "$GGAP_DEINIT");


#############################################################################
##
#F  $GGAP_RESET
##
##  This will be called if $GGAP_API_VERSION is different
##  from the one used in restore.g
##
$GGAP_RESET := function()
  local name, api;

  Info(InfoGGAP, 3, "$GGAP_RESET");

  $GGAP_DEINIT();

  api := $GGAP_DATA.api;
  for name in api do
    UnbindGlobal(name);
  od;
end;


# Call it already!
$GGAP_INIT();

